import flavio
from flavio.classes import Observable, Prediction
import numpy as np
from flavio.physics.edms.common import proton_charges

r"""Functions for neutrinoless $\mu - e$ conversion in different target nuclei"""
def CR_mue(wc_obj, par, nucl):
  r"""Conversion rate independent of the target nucleus"""
  mm = par['m_mu']
  scale = flavio.config['renormalization scale']['mudecays']
  #####overlap integrals and other parameters#####
  # GuV = par['GuV']
  # GdV = par['GdV']
  # GsV = par['GsV']
  pc = proton_charges(par, scale)
  GuS = (pc['gS_u+d'] + pc['gS_u-d']) / 2
  GdS = (pc['gS_u+d'] - pc['gS_u-d']) / 2
  GsS = pc['gS_s']
  D   = par['D ' +nucl]*mm**(5/2)
  Sp  = par['Sp '+nucl]*mm**(5/2)
  Vp  = par['Vp '+nucl]*mm**(5/2)
  Sn  = par['Sn '+nucl]*mm**(5/2)
  Vn  = par['Vn '+nucl]*mm**(5/2)
  GC  = par['GammaCapture '+nucl]
  #####Wilson Coefficients######
  #####Conversion Rate obtained from hep-ph/0203110#####
  wc = wc_obj.get_wc('mue', scale, par, nf_out=3)
  AL = np.sqrt(2)/(4*par['GF']*mm)*wc['Cgamma_emu'].conjugate()
  AR = np.sqrt(2)/(4*par['GF']*mm)*wc['Cgamma_mue']
  gRV = {'u': 4*(wc['CVRR_mueuu'] + wc['CVLR_mueuu']),
         'd': 4*(wc['CVRR_muedd'] + wc['CVLR_muedd']),
         's': 4*(wc['CVRR_muess'] + wc['CVLR_muess'])}
  gLV = {'u': 4*(wc['CVLR_uumue'] + wc['CVLL_mueuu']),
         'd': 4*(wc['CVLR_ddmue'] + wc['CVLL_muedd']),
         's': 4*(wc['CVLR_ssmue'] + wc['CVLL_muess'])}
  gRS = {'u': 4*(wc['CSRR_emuuu'].conjugate() + wc['CSRL_mueuu']),
         'd': 4*(wc['CSRR_emudd'].conjugate() + wc['CSRL_muedd']),
         's': 4*(wc['CSRR_emuss'].conjugate() + wc['CSRL_muess'])}
  gLS = {'u': 4*(wc['CSRL_emuuu'].conjugate() + wc['CSRR_mueuu']),
         'd': 4*(wc['CSRL_emudd'].conjugate() + wc['CSRR_muedd']),
         's': 4*(wc['CSRL_emuss'].conjugate() + wc['CSRR_muess'])}
  lhc = (AR.conjugate()*D + (2*gLV['u'] + gLV['d'])*Vp
         +(gLV['u'] + 2*gLV['d'])*Vn
         + (GuS* gLS['u'] + GdS*gLS['d'] + GsS*gLS['s'])*Sp
         + (GuS*gLS['u'] + GdS*gLS['d'] + GsS*gLS['s'])*Sn)
  rhc = (AL.conjugate()*D + (2*gRV['u'] + gRV['d'])*Vp
         +(gRV['u'] + 2*gRV['d'])*Vn
         + (GuS* gRS['u'] + GdS*gRS['d'] + GsS*gRS['s'])*Sp
         + (GuS*gRS['u'] + GdS*gRS['d'] + GsS*gRS['s'])*Sn)
  return 2*(par['GF']**2)*(abs(lhc)**2 + abs(rhc)**2)/GC


def CR_mueAu(wc_obj, par):
  r"""Conversion rate for $\phantom k^{197}_{79} \mathrm{Au}$"""
  return CR_mue(wc_obj, par, 'Au')
def CR_mueAl(wc_obj, par):
  r"""Conversion rate for $\phantom k^{27}_{13} \mathrm{Al}$"""
  return CR_mue(wc_obj, par, 'Al')
def CR_mueTi(wc_obj, par):
  r"""Conversion rate for $\phantom k^{48}_{22} \mathrm{Ti}$"""
  return CR_mue(wc_obj, par, 'Ti')

CRAu = Observable('CR(mu->e, Au)')
Prediction('CR(mu->e, Au)', CR_mueAu)
CRAu.tex = r"$CR(\mu - e)$ in $\phantom k^{197}_{79} \mathrm{Au}$"
CRAu.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{197}_{79} \mathrm{Au}$"
CRAu.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRAu.tex)

CRAl = Observable('CR(mu->e, Al)')
Prediction('CR(mu->e, Al)', CR_mueAl)
CRAl.tex = r"$CR(\mu - e)$ in $\phantom k^{27}_{13} \mathrm{Al}$"
CRAl.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{27}_{13} \mathrm{Al}$"
CRAl.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRAl.tex)

CRTi = Observable('CR(mu->e, Ti)')
Prediction('CR(mu->e, Ti)', CR_mueTi)
CRTi.tex = r"$CR(\mu - e)$ in $\phantom k^{48}_{22} \mathrm{Ti}$"
CRTi.description = r"Coherent conversion rate of $\mu^-$ to $e^-$ in $\phantom k^{48}_{22} \mathrm{Ti}$"
CRTi.add_taxonomy(r'Process :: muon decays :: LFV decays :: $\mu N \to e N$ :: ' + CRTi.tex)



r"""Functions for $K\to \pi\ell\nu$ decays."""

from math import sqrt
import flavio
from flavio.classes import Observable, Prediction


def get_ff(q2, par, K):
    ff_name = 'K->pi form factor'
    ff_K0 = flavio.classes.AuxiliaryQuantity[ff_name].prediction(par_dict=par, wc_obj=None, q2=q2)
    if K == 'KL' or K == 'KS':
        return ff_K0
    elif K == 'K+':
        # isospin breaking correction for K+->pi0lnu: multiply all FFs by 1+delta
        return {k: (par['K->pi delta_K+pi0'] + 1)*v for k,v in ff_K0.items()}

def get_angularcoeff(q2, wc_obj, par, K, P, lep):
    Jlist = [_get_angularcoeff(q2, wc_obj, par, K, P, lep, nu)
             for nu in ['e', 'mu', 'tau']]
    J = {}
    J['a'] = sum([JJ['a'] for JJ in Jlist])
    J['b'] = sum([JJ['b'] for JJ in Jlist])
    J['c'] = sum([JJ['c'] for JJ in Jlist])
    return J

def _get_angularcoeff(q2, wc_obj, par, K, P, lep, nu):
    GF = par['GF']
    ml = par['m_'+lep]
    mK = par['m_'+K]
    mP = par['m_'+P]
    Vus = flavio.physics.ckm.get_ckm(par)[0,1]
    # renormalization scale is m_rho
    scale = par['m_rho0']
    ms = flavio.physics.running.running.get_ms(par, scale)
    wc = flavio.physics.bdecays.wilsoncoefficients.get_wceff_fccc(wc_obj, par, 'su', lep, nu, ms, scale, nf=3)
    N = 4*GF/sqrt(2)*Vus
    ff = get_ff(q2, par, K)
    h = flavio.physics.bdecays.angular.helicity_amps_p(q2, mK, mP, ms, 0, ml, 0, ff, wc, N)
    J = flavio.physics.bdecays.angular.angularcoeffs_general_p(h, q2, mK, mP, ms, 0, ml, 0)
    return J


def dGdq2(J):
    return 2 * (J['a'] + J['c']/3.)

def dBRdq2(q2, wc_obj, par, K, P, lep):
    ml = par['m_'+lep]
    mK = par['m_'+K]
    mP = par['m_'+P]
    if q2 < ml**2 or q2 > (mK-mP)**2:
        return 0
    tauK = par['tau_'+K]
    J = get_angularcoeff(q2, wc_obj, par, K, P, lep)
    if P == 'pi0':
        # factor of 1/2 for neutral pi due to pi = (uubar-ddbar)/sqrt(2)
        return tauK * dGdq2(J) / 2.
    if K == 'K+':
        deltaEM = par['K+' + lep + '3 delta_EM'] # e.g. 'K+e3 delta_EM'
    elif K == 'KL' or K == 'KS':
        deltaEM = par['K0' + lep + '3 delta_EM'] # e.g. 'K+e3 delta_EM'
    return tauK * dGdq2(J) * (1 + deltaEM)**2

def BR_binned(q2min, q2max, wc_obj, par, K, P, lep):
    def integrand(q2):
        return dBRdq2(q2, wc_obj, par, K, P, lep)
    return flavio.math.integrate.nintegrate(integrand, q2min, q2max)

def BR_tot(wc_obj, par, K, P, lep):
    mK = par['m_'+K]
    mP = par['m_'+P]
    ml = par['m_'+lep]
    q2max = (mK-mP)**2
    q2min = ml**2
    return BR_binned(q2min, q2max, wc_obj, par, K, P, lep)

def BR_tot_function(K, P, lep):
    return lambda wc_obj, par: BR_tot(wc_obj, par, K, P, lep)


# Observable and Prediction instances

_tex = {'e': 'e', 'mu': '\mu', 'l': r'\ell'}
_tex_br = {'dBR/dq2': r'\frac{d\text{BR}}{dq^2}', 'BR': r'\text{BR}', '<BR>': r'\langle\text{BR}\rangle'}
_args = {'dBR/dq2': ['q2'], 'BR': None, '<BR>': ['q2min', 'q2max']}
_hadr = {
'KL->pi': {'tex': r"K_L\to \pi^+", 'K': 'KL', 'P': 'pi+', },
'KS->pi': {'tex': r"K_S\to \pi^+", 'K': 'KS', 'P': 'pi+', },
'K+->pi': {'tex': r"K^+\to \pi^0", 'K': 'K+', 'P': 'pi0', },
}

for l in ['e', 'mu', 'l']:
    for M in _hadr.keys():
        _process_tex = _hadr[M]['tex']+_tex[l]+r"^+\nu_"+_tex[l]
        _process_taxonomy = r'Process :: $s$ hadron decays :: Semi-leptonic tree-level decays :: $K\to P\ell\nu$ :: $' + _process_tex + r"$"

        _obs_name = "BR("+M+l+"nu)"
        _obs = Observable(_obs_name)
        _obs.set_description(r"Total branching ratio of $" + _process_tex + r"$")
        _obs.tex = r"$\text{BR}(" + _process_tex + r")$"
        _obs.add_taxonomy(_process_taxonomy)
        Prediction(_obs_name, BR_tot_function(_hadr[M]['K'], _hadr[M]['P'], l))



import numpy as np
import json
import pkgutil
from flavio.classes import Parameter
from flavio.statistics.probability import MultivariateNormalDistribution


FFs = ["f+", "fT", "f0"]
ai = ["a0", "a1", "a2"]
ff_a = [(ff, a) for ff in FFs for a in ai]
a_ff_string = [a + '_' + ff for ff in FFs for a in ai]


tex_a = {'a0': 'a_0', 'a1': 'a_1', 'a2': 'a_2', }
tex_ff = {'f+': 'f_+', 'fT': 'f_T', 'f0': 'f_0', }


def get_ffpar(filename):
    f = pkgutil.get_data('flavio.physics', filename)
    data = json.loads(f.decode('utf-8'))
    central = np.array([data['central'][ff].get(a, np.nan) for ff, a in ff_a])
    unc = np.array([data['uncertainty'][ff].get(a, np.nan) for ff, a in ff_a])
    corr = np.array([[data['correlation'][ff1 + ff2].get(a1 + a2, np.nan) for ff1, a1 in ff_a] for ff2, a2 in ff_a])
    # delete the parameter a0_f0, which is instead fixed
    # using the exact kinematical relation f0(0) = f+(0)
    pos_a0_f0 = ff_a.index(('f0', 'a0'))
    central = np.delete(central, [pos_a0_f0])
    unc = np.delete(unc, [pos_a0_f0])
    corr = np.delete(corr, [pos_a0_f0], axis=0)
    corr = np.delete(corr, [pos_a0_f0], axis=1)
    return [central, unc, corr]


def load_parameters(filename, process, constraints):
    implementation_name = process + ' BSZ'
    parameter_names = [implementation_name + ' ' + coeff_name for coeff_name in a_ff_string]
    # a0_f0 is not treated as independent parameter!
    parameter_names.remove(implementation_name + ' a0_f0')
    for parameter_name in parameter_names:
        try:  # check if parameter object already exists
            p = Parameter[parameter_name]
        except KeyError:  # if not, create a new one
            p = Parameter(parameter_name)
            # get LaTeX representation of coefficient and form factor names
            _tex_a = tex_a[parameter_name.split(' ')[-1].split('_')[0]]
            _tex_ff = tex_ff[parameter_name.split(' ')[-1].split('_')[-1]]
            p.tex = r'$' + _tex_a + r'^{' + _tex_ff + r'}$'
            p.description = r'BSZ form factor parametrization coefficient $' + _tex_a + r'$ of $' + _tex_ff + r'$'
        else:  # if parameter exists, remove existing constraints
            constraints.remove_constraint(parameter_name)
    [central, unc, corr] = get_ffpar(filename)
    constraints.add_constraint(parameter_names,
        MultivariateNormalDistribution(central_value=central, covariance=np.outer(unc, unc)*corr))



# Resonance masses used in arXiv:1811.00983
resonance_masses_gkvd = {
    'B->K': {
        'm0': 5.630,
        'm+': 5.412,
    },
    'B->pi': {
        'm0': 5.540,
        'm+': 5.325,
    },
    'B->D': {
        'm0': 6.420,
        'm+': 6.330,
    },
}


def gkvd_load(version, fit, transitions, constraints):
    """Load the form factor parameters given in arXiv:1811.00983"""
    for tr in transitions:
        for m, v in resonance_masses_gkvd[tr].items():
            constraints.set_constraint('{} BCL {}'.format(tr, m), v)
        filename = 'data/arXiv-1811-00983{}/{}_{}.json'.format(version, tr.replace('->', ''), fit)
        load_parameters(filename, tr, constraints)
